from typing import Any

from simudator.core.module import Module
from simudator.core.signal import Signal


class Demux(Module):
    """
    A general demux that allows an arbitrary amount of outputs to map
    to a single input. The output side is controlled by the control
    signal 'from_demux_s'.
    """

    def __init__(self, from_demux: Signal, bit_length: int,
                 input: Signal, outputs: list[Signal] = [],
                 name="demux", value=0)  -> None:
        # Name
        super().__init__(name)

        # Signals
        self.from_demux_s = from_demux
        self.outputs = outputs
        self.input_s = input

        # mask and bit_length
        self.bit_length = bit_length
        self.mask = 2**self.bit_length -1

        # Value to be read/written
        self.value = value

        # Set the demux as destination of its input signal
        self.input_s.add_destination(self)

    def update_register(self) -> None:
        """
        Read the input value and update the value of the demux.
        """
        input_value = self.input_s.get_value()
        self.value = input_value & self.mask


    def output_register(self):
        """
        Read which signal to write to from the control signal
        'from_demux_s' and forward the input to that output signal.
        """
        output_index = self.from_demux_s.get_value()
        self.outputs[output_index].update_value(self.value)

    def update_logic(self):
        """
        The demux has no logic
        """
        pass

    def get_state(self) -> dict[str: Any]:
        """
        Returns a dict of the demux state.
        """
        state = super().get_state()
        state["value"] = self.value
        state["bit_length"] = self.bit_length
        state["mask"] = self.mask

        return state

    def set_state(self, state: dict) -> None:
        """
        Sets the name and value of the demux.
        """
        super().set_state(state)
        self.value = state["value"]
        self.bit_length = state["bit_length"]
        self.mask = state["mask"]
