from typing import Any

from simudator.core.module import Module
from simudator.core.signal import Signal


class Mux(Module):
    """
    A general mux that allows an arbitrary amount of input to map
    to a single output. The input side is controlled by the control
    signal 'to_mux_s'.
    """

    def __init__(self,
                 to_mux: Signal,
                 bit_length: int,
                 output: Signal,
                 inputs: list[Signal] = [],
                 name="mux",
                 value=0
                 ) -> None:
        print("in: ", inputs)
        # Name
        super().__init__(name)

        # Signals
        self.to_mux_s = to_mux
        self.output_s = output
        self.inputs = inputs

        # mask and bit_length
        self.bit_length = bit_length
        self.mask = 2**self.bit_length -1

        # Value to be read/written
        self.value = value

        # Set mux as destination of all input signals
        if self.inputs: # make sure inputs is not empty
            for signal in self.inputs:
                signal.add_destination(self)

    def update_register(self) -> None:
        """
        Read which signal to read from the control signal
        'to_mux_s' and forward that signal to the output.
        """
        input_index = self.to_mux_s.get_value()
        input_value = self.inputs[input_index].get_value()
        self.value = input_value & self.mask

    def output_register(self):
        """
        Output the value of the mux to its output.
        """
        self.output_s.update_value(self.value)

    def update_logic(self):
        """
        The mux has no logic.
        """
        pass

    def get_state(self) -> dict[str: Any]:
        """
        Returns a dict of the mux state.
        """
        state = super().get_state()
        state["value"] = self.value
        state["bit_length"] = self.bit_length
        state["mask"] = self.mask

        return state

    def set_state(self, state: dict) -> None:
        """
        Sets the name and value of the mux.
        """
        super().set_state(state)
        self.value = state["value"]
        self.bit_length = state["bit_length"]
        self.mask = state["mask"]

    def print_module(self) -> None:
        print(self.name, "\n-----",
              "\nvalue: ", self.value,
              "\nbit length: ", self.bit_length,
              "\nmask: ", self.mask,
              )
