import os
from collections.abc import Callable
from typing import Any

from simudator.core.breakpoint import Breakpoint
from simudator.core.breakpoint_lambda import LambdaBreakpoint
from simudator.core.breakpoint_memory import MemoryBreakpoint
from simudator.core.breakpoint_state import StateBreakpoint
from simudator.core.module import Module
from simudator.core.modules.memory import Memory
from simudator.processor.mia.modules.bus import Bus
from simudator.processor.mia.modules.micro_memory import MicroMemory

from .signal import Signal


class Processor:
    """
    Main class for controlling the processor.
    Uses modules and signals to simulate processor components behaviour.
    """

    def __init__(self):
        self.modules: dict[str, Module] = dict()
        self.signals = []
        self.clock = 0
        self.update_queue = []
        self.module_history: list[dict[str, dict[str, Any]]] = []
        self.signal_history: list[list] = []    # TODO: Is this needed?
        self.breakpoint_id_counter = 1
        self.breakpoints: dict[int, Breakpoint] = {}
        self.breakpoint_reached = False
        self.last_breakpoint = None
        self.cycles_to_save = 1000
        self.removed_cycles = 0
        self.max_line_len = 170
        self.line_seperator = "-"
        self.is_stopped = False
        # TODO: keeping track of what pieces of info not to show
        # show not be done at the processor level.
        # Maybe implemenet a 'get_pretty_print_state' at module
        # level?
        self.ignore_keys = [
                "bit_length", "mask", "increment", "read_from_bus",
                "read_from_uADR", "decrement_by_one", "bus_id"
                ]
        self.lambdas: dict[str, Callable[..., bool]] = {}

    def do_tick(self) -> None:
        """
        Does one clock-cycle of the processor by running each modules
        tick function, then handels the following updates.
        """
        if len(self.module_history) > self.clock - self.removed_cycles:
            # If a previous stored cycle has been loaded, discard
            # all stored cycles from that cycle and onward in
            # the history of saved cycles
            self.module_history = self.module_history[0:self.clock]
            self.signal_history = self.signal_history[0:self.clock]

        self.save_cycle()
        self.clock += 1

        for module in self.modules.values():
            module.update_register()

        for module in self.modules.values():
            module.output_register()

        while self.update_queue:
            module = self.update_queue.pop(0)
            module.update_logic()

    def get_current_instrution(self):
        """
        Return the current instruction. Useful for pipeline diagram.
        """
        raise NotImplemented

    def run_continuously(self):
        """
        Runs the processor until it halts.
        """
        while not self.should_halt() and not self.is_stopped:
            self.do_tick()

            # Print all reached breakpoints then exit the loop
            self.breakpoint_reached = False
            for _, bp in self.breakpoints.items():
                #TODO: Can make this more efficient by only checking enabled breakpoints
                if bp.is_break() and bp.is_enabled:
                    self.breakpoint_reached = True
                    self.last_breakpoint = bp

            if self.breakpoint_reached:
                return

    def should_halt(self) -> bool:
        return False

    def stop(self) -> None:
        self.is_stopped = True

    def unstop(self) -> None:
        self.is_stopped = False

    def reset(self):
        """
        Resets all values in the processor and
        then does one propogation to the signals.
        Is run before running the processor to propogate values in
        module out into signals of the system.
        """
        self.clock = 0
        self.module_history.clear()
        self.signal_history.clear()

        for module in self.modules.values():
            module.reset()
            module.output_register()

        while self.update_queue:
            module = self.update_queue.pop(0)
            module.update_logic()


    def add_modules_to_update(self, module: Module) -> None:
        """
        Queues module to be updated at the end of clock cycle.
        """
        if module not in self.update_queue:
            self.update_queue.append(module)

    def add_module(self, module: Module) -> None:
        """
        Add module into the processor.
        """
        self.modules[module.name] = module

    def get_module(self, name: str) -> Module:
        """
        Get module with specific name.
        """
        return self.modules[name]

    def get_modules(self) -> list[Module]:
        """
        Get list of all modules.
        """
        return list(self.modules.values())

    def add_signals(self, signals: list[Signal]) -> None:
        """
        Add signals into the processor.
        """
        self.signals += signals

    def get_clock(self) -> int:
        """
        Get current clockcycle.
        """
        return self.clock

    def set_clock(self, value: int) -> int:
        """
        Set current clockcycle.
        """
        self.clock = value

    def save_cycle(self) -> None:
        """
        Save the current state for all the modules and signals in the
        processor to the processor's internal history of all clock cycles.
        """
        module_states = {}

        # Only save a specified number of cycles,
        # saving every cycle can easily eat all ram
        if len(self.module_history) > self.cycles_to_save:
            self.module_history.pop(0)
            self.removed_cycles += 1

        for module in self.modules.values():
            module_states[module.name] = module.get_state()
        self.module_history.append(module_states)

    def load_cycle(self, cycle: int) -> None:
        """
        Load the state of all modules as they were at the specified
        clock cycle. This does not erase the history of states for the
        later cycles.
        """

        cycle_index = cycle - self.removed_cycles

        if cycle_index < 0:
            raise ValueError("The cycle to be loaded is not saved")

        try:
            module_states = self.module_history[cycle_index]
        except IndexError:
            raise IndexError

        for module_name, module_state in module_states.items():
            self.modules[module_name].set_state(module_state)

        self.clock = cycle

        for module in self.modules.values():
            module.output_register()

        while self.update_queue:
            module = self.update_queue.pop(0)
            module.update_logic()

    def load_state_from_file(self, file_path) -> None:
        """
        Loads states for modules from a file.
        """
        file = open(file_path)

        # Ensure we are at the start of the file
        file.seek(0, os.SEEK_END)   # Go to end of file
        if file.tell():             # If current pos != 0
             file.seek(0)           # Return to the start of the file
        else:
            raise ValueError        # Else raise error


        module_name = None
        module_str = ""
        # Go through each line in the file
        lines = file.readlines()
        for line in lines:

            # If no name currently saved then get name from current line
            if module_name is None:
                module_name = line.partition(":")[0]

            # if already has name get line content
            else:
                # if not empty line, then save content
                if line.strip():
                    module_str += line

                # if line is empty then give saved content to name file
                else:
                    module = self.get_module(module_name)
                    module.load_from_str(module_str)
                    module_str = ""
                    module_name = None

        # give module anything that is left
        if module_name and module_str:
            module = self.get_module(module_name)
            module.load_from_str(module_str)

    def save_state_to_file(self, file_path: str) -> None:
        """
        Saves the current state of the cpu to a given file.
        Will overwrite the given file.
        """
        file = open(file_path, "w")
        file.write("")
        file.close()

        for module in self.modules.values():
            module.save_state_to_file(file_path)

    def pretty_print(self) -> None:
        self.pretty_print_verbose(self.ignore_keys)

    def pretty_print_verbose(self, ignore_keys = []) -> None:
        """
        Prints the most relevant information about each module
        compact and pretty. Fields of modules can be ignored with the
        optional argument 'ignore_keys'.
        """
        # TODO: ignore keys per module and not for all modules
        memory_modules = []
        other_modules = []

        # TODO: remove isinstance(module, micro_memory)
        for module in self.modules.values():
            if not isinstance(module, Bus):
                if isinstance(module, Memory) \
                    or isinstance(module, MicroMemory):
                    memory_modules.append(module)
                else:
                    other_modules.append(module)

        # sort the modules by name to ensure that they appear in the
        # same order
        memory_modules.sort(key=lambda x: x.name, reverse=True)
        other_modules.sort(key=lambda x: x.name, reverse=True)

        # specify which keys to ignore from the modules 'get_state'
        # function
        # TODO: ignore fields per module, this will ignore 'mask'
        # for every module

        module_to_line_length = {}
        # get the longest line length for all other_modules
        for module in other_modules:
            line_len = module.get_longest_line_len(ignore_keys) + 3 #+3 for padding

            # TODO: what to do if two or more modules has the same name
            module_to_line_length[module] = line_len

        groups = self.group_pp_modules(module_to_line_length)

        print(self.line_seperator * self.max_line_len)

        # print each group seperate
        for group in groups:
            self.pretty_print_names(group)

            # Build a new string containing the information to show
            # the user
            for row in range(self.get_most_fields(group, ignore_keys)):
                string = ""
                total_padding = 0

                # Get the information from each module
                for module in group:
                    module_state = module.get_state()
                    keys = self.fields_to_list(module)

                    # Get the keys that we want to print to the user
                    # If this is not done, the keys we want to print
                    # could end up on an index that is higher than the
                    # number of rows printed and will therefore be missed
                    real_keys = [key for key in keys
                                 if key not in ignore_keys]

                    # Dont go out of index
                    # Needed since one module might want to print 4
                    # fields and another only 1
                    if row < len(real_keys) and real_keys[row] \
                    not in ignore_keys:
                        string += real_keys[row] + ": " \
                        + str(module_state[real_keys[row]])

                    # padd the string so each string has the same
                    # length
                    total_padding += module_to_line_length[module]
                    string = string.ljust(total_padding)
                    # replace last to chars with a separator and padding
                    string = string[0:-2] + "| "

                print(string)
            print(self.line_seperator * self.max_line_len)

        for memory_module in memory_modules:
            self.pretty_print_memory(memory_module)

    def pretty_print_memory(self, module: Module) -> None:
        print(module.name)
        print(self.line_seperator * self.max_line_len)

        longest_line_len = module.get_longest_line_len()
        #longest_line_len = self.get_longest_memory_value(module.memory)
        string = ""
        last_mem_len = module.get_largest_mem_adr()

        for i, value in enumerate(module.memory):

            # create a new string containg the adress and the value
            # of the memory adress formattet to fit with the largest
            # adress and largest memory value
            new_row = ((str(i).ljust(last_mem_len)
                       + ": "
                       + str(value)).ljust(longest_line_len
                       + last_mem_len + 3)
                       + "|")

            # only add the string if there is space for it, else
            # print the string and start a new
            if len(string + new_row) + 1 > self.max_line_len:
                print(string)
                string = new_row
            else:
                # First iteration string will be
                # empty and should not be padded
                if string:
                    string += " " + new_row
                else:
                    string = new_row
        print(string)
        print(self.line_seperator * self.max_line_len)
        print()

    def pretty_print_names(self,
                           module_to_line_length: dict[Module, int]
                           ) -> None:
        """
        Prints the name of the modules in one row with enough
        added spacing between the names so that the modules
        longest field will have space to be printed below.
        """
        name_string = ""
        total_len = 0
        for module in module_to_line_length:
            name_string += module.name
            total_len += module_to_line_length[module]
            name_string = name_string.ljust(total_len)
            name_string = name_string[0:-2] + "| "

        print(name_string)

    def group_pp_modules(self,
                         module_to_line_length: dict[Module, int]
                         ) -> list[dict[Module, int]]:
        """
        Groups the modules to be pretty printed into groups with a
        total line length lower than 'self.max_line_len'.
        """

        groups = [{}]
        group_index = 0
        line_len = 0
        for module in module_to_line_length:

            # Make sure the line is not to long
            if line_len + module_to_line_length[module] < self.max_line_len:
                line_len += module_to_line_length[module]
                groups[group_index][module] = module_to_line_length[module]

            # If it would have been, start a new group
            else:
                groups.append({})
                group_index += 1
                groups[group_index][module] = module_to_line_length[module]
                line_len = module_to_line_length[module]
        return groups

    def fields_to_list(self, module: Module, ignore_keys=[]) -> list[str]:
        """
        Return a list containing all the keys for a module except the name.
        Optinal argument to ignore specific keys.
        """
        return [key for key in module.get_state() if key != "name"
                and key not in ignore_keys]

    def get_most_fields(self, modules: dict[Module: int],
                        ignore_keys=[]) -> int:
        """
        Returns the most number of a fields a module has. Can optionally
        ignore keys.
        """
        fields = 0
        for module in modules:
            module_fields = len(self.fields_to_list(module, ignore_keys))
            if module_fields > fields:
                fields = module_fields

        return fields

    def add_state_breakpoint(self, module_name: str, state_name: str,
                       value: Any) -> None:

        if module_name not in self.modules:
            raise ValueError(f"No module named {module_name}")

        module = self.modules[module_name]

        if state_name not in module.get_state():
            raise ValueError(f"No state named {state_name} in "
                             f"module {module_name}")

        bp = StateBreakpoint(module, state_name, value)
        self.breakpoints[self.breakpoint_id_counter] = bp
        self.breakpoint_id_counter += 1

    def remove_breakpoint(self, id: int) -> bool:
        if id in self.breakpoints:
            del self.breakpoints[id]
            return True
        return False

    def print_breakpoints(self) -> None:
        if not self.breakpoints:
            print("There are no breakpoints.")
        else:
            for bp_id, bp in self.breakpoints.items():
                print(f"BP {bp_id}: {bp}")

    def add_memory_breakpoint(self, module_name:str, adress: int, value: Any) -> None:

        if module_name not in self.modules:
            raise ValueError(f"No module named {module_name}")

        module = self.modules[module_name]
        if "memory" not in module.get_state():
            raise ValueError(f"Module {module_name} is not a memory.")

        bp = MemoryBreakpoint(module, adress, value)
        self.breakpoints[self.breakpoint_id_counter] = bp
        self.breakpoint_id_counter += 1

    def get_breakpoint_lambdas(self) -> list[str]:
        return list(self.lambdas.keys())

    def add_lambda_breakpoint(self, lambda_name: str, **kwargs) -> None:
        # Substitute any module names for the module instances
        # This is to be able to externally create general lambda breakpoints
        # that can take any values as arguments while also having a way to take
        # modules as arguments without having direct access to them
        for key in kwargs.keys():
            value = kwargs[key]
            if value in self.modules:
                kwargs[key] = self.modules[value]

        if lambda_name not in self.lambdas:
            raise ValueError(f"No lambda named {lambda_name}.")

        lambda_func = self.lambdas[lambda_name]
        bp = LambdaBreakpoint(lambda_func, **kwargs)
        self.breakpoints[self.breakpoint_id_counter] = bp
        self.breakpoint_id_counter += 1

    def set_enabled_breakpoint(self, bp_id: int, is_enabled: bool) -> None:
        self.breakpoints[bp_id].set_enabled(is_enabled)
