from qtpy.QtWidgets import (
    QHBoxLayout,
    QListWidget,
    QListWidgetItem,
    QPushButton,
    QStyle,
    QVBoxLayout,
    QWidget,
)

from simudator.core.breakpoint import Breakpoint
from simudator.core.processor import Processor


class BreakpointWindow(QWidget):
    def __init__(self, cpu: Processor):
        super().__init__()
        self.cpu = cpu

        # Make layouts
        layout = QHBoxLayout()
        button_layout = QVBoxLayout()

        # Make buttons for editing breakpoints
        self.toggle_enabled_button = QPushButton("Toggle Enabled")
        self.toggle_enabled_button.clicked.connect(self.toggleEnabledClicked)
        button_layout.addWidget(self.toggle_enabled_button)

        self.remove_button = QPushButton("Remove")
        self.remove_button.clicked.connect(self.removeItemClicked)
        button_layout.addWidget(self.remove_button)
        layout.addLayout(button_layout)

        # Add list of breakpoints
        self.breakpoint_list = QListWidget()
        layout.addWidget(self.breakpoint_list)

        self.setLayout(layout)

        # Update so we fill the list with breakpoints
        self.update()

    def update(self) -> None:
        """
        Updates list with breakpoints by removing the current items
        and then filling with new once according to the cpu's content.
        """
        # Remove so we don't have dubble of all items
        self.removeAllItems()

        # Get icons for showing if enabled
        i_icon = self.style().standardIcon(QStyle.SP_DialogApplyButton)
        stop_icon = self.style().standardIcon(QStyle.SP_DialogCancelButton)

        # Add an list item for each breakpoint in cpu
        for bp_id, breakpoint in self.cpu.breakpoints.items():
            bp_str = str(bp_id) + ": " + breakpoint.__str__()
            bp_item = QListWidgetItem(bp_str)

            # Add icon depending on enabled
            if breakpoint.is_enabled:
                bp_item.setIcon(i_icon)
            else:
                bp_item.setIcon(stop_icon)
            self.breakpoint_list.addItem(bp_item)

    def removeAllItems(self) -> None:
        """
        Removes all list items from the breakpoint list.
        """
        for _ in range(self.breakpoint_list.count()):
            self.breakpoint_list.takeItem(0)

    def toggleEnabledClicked(self) -> None:
        """
        Called when the toggle enabled buttons is pressed.
        Toggles the selected breakpoint so it is enabled or disabled.
        """
        current_bp = self.getSelectedItem()

        if current_bp is None:
            return

        current_bp.set_enabled(not current_bp.is_enabled)
        self.update()

    def removeItemClicked(self) -> None:
        """
        Called when the remove button is preseed.
        Will remove the selected breakpoint from the cpu's breakpoint list.
        """
        bp_id = self.getSelectedItemId()

        # Do nothing if id is None
        if bp_id is None:
            return

        del self.cpu.breakpoints[bp_id]
        self.update()

    def getSelectedItemId(self) -> int:
        """
        Gives the id of the selected breakpoint.
        Used to fetch the corresponding breakpoint from the cpu.
        """
        current_bp_item = self.breakpoint_list.currentItem()

        # Return None if list is empty
        if current_bp_item is None:
            return

        return int(current_bp_item.text()[0])

    def getSelectedItem(self) -> Breakpoint:
        """
        Returns the selected breakpoint.
        """
        bp_id = self.getSelectedItemId()

        # Do nothing if there are no breakpoints or if id is nonetype
        if bp_id is None or bp_id == -1:
            return
        return self.cpu.breakpoints[bp_id]
