from qtpy.QtWidgets import QGraphicsItem, QGraphicsScene

from simudator.core.processor import Processor
from simudator.gui.color_scheme import ColorScheme
from simudator.gui.module_graphics_item.module_graphics_item import ModuleGraphicsItem
from simudator.gui.signal_graphics_item import SignalGraphicsItem


class CpuGraphicsScene(QGraphicsScene):
    """
    This class creates the graphics scene for visualising the processor.
    It takes each module representation as a ModuleGraphicsItem
    and handels mouse inputs for interacting with these graphicsitems.
    Can create a default layout on creation and can save/load new layouts.
    """

    MODULE_SPACEING = 100

    def __init__(self, cpu: Processor):
        super().__init__()
        self.cpu = cpu
        self.module_graphics_items = dict()
        self.signal_graphics_items = []
        self.setBackgroundBrush(ColorScheme.Window)

    def resetSignals(self) -> None:
        """
        Resets all graphical signals to their default visual representation
        when initialised.
        """
        for graphics_Signal in self.signal_graphics_items:
            graphics_Signal.reset()

    def addModuleGraphicsItem(self, graphics_item: ModuleGraphicsItem) -> None:
        """
        Takes a ModuleGraphicsItem and adds it to the scene.
        Will give it a new position according to the deafult layout.
        """
        self.module_graphics_items[graphics_item.name] = graphics_item
        self.placeModuleGraphicsItemDefault(graphics_item)

    def placeModuleGraphicsItemDefault(self, graphics_item: ModuleGraphicsItem) -> None:
        """
        Places a module graphics items at a position where the x value
        is equal to the y value. Used to place all graphics items in a
        diagonal.
        """
        placement = len(self.module_graphics_items) * self.MODULE_SPACEING
        graphics_item.setPos(placement, placement)
        self.addItem(graphics_item)

    def replaceModuleGraphicsItem(
        self, graphics_item: ModuleGraphicsItem, pos_x: int, pos_y: int
    ) -> None:
        """
        Changes the postions of an existing modules graphics item.
        """
        graphics_item.setPos(pos_x * self.MODULE_SPACEING, pos_y * self.MODULE_SPACEING)

    def updateGraphicsItems(self):
        """
        Used to update graphicsitems when modules in the processor has chnaged values.
        """
        for graphics_item in self.module_graphics_items.values():
            graphics_item.update()

    def addAllSignals(self) -> None:
        """
        Instantiates signals between all matching ports of all modules.
        """
        # Map which ports that are connected using their signals' names
        # (Every port keeps a reference to a simulation signal which has
        # a name)
        signal_to_ports = {}
        for module_w in self.module_graphics_items.values():
            ports = module_w.getPorts()
            for port in ports:
                s_name = port.getSignalName()
                if s_name not in signal_to_ports:
                    signal_to_ports[s_name] = [port]
                else:
                    signal_to_ports[s_name].append(port)

        # Create a signal graphics_item for each bundle of connected ports
        for module_graphics_items in signal_to_ports.values():
            # TODO: Implement signals that go to multiple destinations
            if len(module_graphics_items) == 2:
                port_1 = module_graphics_items[0]
                port_2 = module_graphics_items[1]
                signal_w = SignalGraphicsItem(port_1, port_2)
                self.signal_graphics_items.append(signal_w)
                self.addItem(signal_w)
                port_1.moved.connect(signal_w.move)
                port_2.moved.connect(signal_w.move)
                port_1.toggled.connect(signal_w.toggleVisibility)
                port_2.toggled.connect(signal_w.toggleVisibility)

    def getModulesGraphicsItems(self) -> list[ModuleGraphicsItem]:
        return list(self.module_graphics_items.values())

    def moduleGraphicsItemsDict(self) -> dict[str, ModuleGraphicsItem]:
        return self.module_graphics_items

    def getSignalGraphicsItems(self) -> list[SignalGraphicsItem]:
        return self.signal_graphics_items

    def mousePressEvent(self, event):
        super().mousePressEvent(event)

    def load_layout_from_file(self, file_path: str) -> None:
        """
        Loads a layout for a processor from a saved filed.
        """
        file = open(file_path)

        graphics_item_name = None
        graphics_item_str = ""

        # Go through each line in the file
        for line in file.readlines():
            # If no name currently saved then get name from current line
            if graphics_item_name is None:
                graphics_item_name = line.partition(":")[0]

            # if already has name get line content
            else:
                # if not empty line, then save content
                if line.strip():
                    graphics_item_str += line

                # if line is empty then give saved content to name file
                else:
                    graphics_item = self.get_module(graphics_item_name)
                    try:
                        graphics_item.load_state_from_str(graphics_item_str)
                    except Exception as exc:
                        raise exc

                    # set back to empty
                    graphics_item_str = ""
                    graphics_item_name = None

        # give module anything that is left
        if graphics_item_name and graphics_item_str:
            graphics_item = self.get_module(graphics_item_name)
            graphics_item.load_state_from_str(graphics_item_str)

    def save_layout_to_file(self, file_path: str) -> None:
        """
        Saves the positions and visibility of graphicsitems in the scene (this
        includes module and signal representations) to a file.
        """
        layout_str = ""
        for graphics_item in self.module_graphics_items.values():
            layout_str += graphics_item.save_state_as_str() + "\n"

        file = open(file_path, "w")
        file.write(layout_str)
        file.close()

    def setAllSignalsVisibility(self, is_signals_visible: bool) -> None:
        for item in self.signal_graphics_items:
            item.setVisible(is_signals_visible)

    def setPortNamesVisibility(self, is_ports_visible: bool) -> None:
        for item in self.module_graphics_items.values():
            for port in item.ports:
                port.setNameVisibility(is_ports_visible)

    def setLayoutLock(self, is_layout_locked: bool) -> None:
        """
        Toggles the layout lock making it so items in the scene can not be moved.
        """

        for item in self.module_graphics_items.values():
            item.setFlag(QGraphicsItem.ItemIsMovable, not is_layout_locked)
            item.setLocked(is_layout_locked)

        for item in self.signal_graphics_items:
            item.setFlag(QGraphicsItem.ItemIsMovable, not is_layout_locked)

            # We use this value so lines in the signal can not be moved or edited
            item.is_locked = is_layout_locked
