import ast

from qtpy.QtCore import Slot
from qtpy.QtWidgets import (
    QGraphicsRectItem,
    QGraphicsSimpleTextItem,
    QTextEdit,
    QVBoxLayout,
    QWidget,
)

from simudator.core.modules import Memory
from simudator.gui.color_scheme import ColorScheme as CS
from simudator.gui.module_graphics_item.memory_graphic import MemoryGraphicsItem
from simudator.gui.orientation import Orientation
from simudator.gui.port_graphics_item import PortGraphicsItem
from simudator.processor.mia.gui.mia_memory_content_dialog import MiaMemoryContentDialog


class MiaMemoryWindow(QWidget):
    """
    Widget for showing content of memory
    """

    ROW_LENGTH = 4

    def __init__(self, memory_module: Memory):
        super().__init__()
        self.module = memory_module

        self.text = QTextEdit("")
        layout = QVBoxLayout()
        layout.addWidget(self.text)
        self.setLayout(layout)

        self.update()

    def update(self):
        memory_str = ""
        for adress, value in enumerate(self.module.memory):
            # Add adress and content to string
            # Make sure its unifrom lenght so rows are consistent
            memory_str += f"0x{adress:02x}" + ": " + f"0x{value:04x}" + "   " + "\t"

            # Make new line when we reach end of row
            if adress % self.ROW_LENGTH == self.ROW_LENGTH - 1:
                memory_str += "\n"

        self.text.setText(memory_str)


class MiaMemoryGraphicsItem(MemoryGraphicsItem):
    """
    Graphics module for a Memory module.
    """

    def __init__(self, memory_module: Memory, **kwargs):
        super().__init__(memory_module, **kwargs)
        self.memory_window = None

    def draw_graphics_item(self) -> None:
        # Same as normal memory but no control signal

        # Create base backgrond rect
        self.baserect = QGraphicsRectItem(0, 0, self._width, self._height, self)
        self.baserect.setPen(CS.ModuleOutline)
        self.baserect.setBrush(CS.ModuleBackground)

        # Make ports
        self.input = PortGraphicsItem(
            self.module.get_input_signal(), Orientation.RIGHT, self
        )
        self.input.setPos(self._width, self._height / 8)

        self.output = PortGraphicsItem(
            self.module.get_output_signal(), Orientation.RIGHT, self
        )
        self.output.setPos(self._width, self._height / 4)

        self.adress = PortGraphicsItem(
            self.module.get_adress_signal(), Orientation.DOWN, self
        )
        self.adress.setPos(self._width / 4, self._height)

        self.ports = [self.input, self.output, self.adress]

        # Create name lable
        name_text = QGraphicsSimpleTextItem(self.state["name"], self)
        name_text.setPos(self._width / 2 - len(self.state["name"]) * self._CHAR_LEN, 0)

        # Create adress lable
        self.adress_text = QGraphicsSimpleTextItem(
            "adress: " + str(self.state["current_adress"]), self
        )
        self.adress_text.setPos(self._width / 20, self._height / 8)

    def showMemoryContents(self) -> None:
        """
        Create and show a MemoryWindow that displays the contents
        of the memory module associated with this graphics item.
        """
        self.memory_window = MiaMemoryWindow(self.module)
        self.memory_window.show()

    def memoryBreakpointDialog(self) -> None:
        """
        Opens dialog window for user to create a breakpoint.
        """
        self.memory_br_dialog = MiaMemoryContentDialog(self.module)
        self.memory_br_dialog.accepted.connect(self.memoryBreakpointAccepted)

    @Slot(str, str, str)
    def memoryBreakpointAccepted(
        self, module_name: str, adress: str, value: str
    ) -> None:
        """
        Same as prent function but also prases data so it is hexadecimal.
        """
        try:
            parsed_adress = int(adress, 16)
            parsed_value = ast.literal_eval(value)
            self.new_memory_breakpoint_signal.emit(
                module_name, str(parsed_adress), str(parsed_value)
            )
        except SyntaxError as e:
            self.errorMessageWidget.showMessage(str(e))
        except ValueError:
            self.errorMessageWidget.showMessage(
                "You must enter a hexadecimal" "number preceeded by '0x' (e.g." "0xc3)."
            )

    @Slot(str, str, str)
    def editMemoryAccepted(self, module_name: str, adress: str, value: str) -> None:
        """
        Same as prent function but also prases data so it is hexadecimal.
        """
        try:
            parsed_adress = int(adress, 16)
            parsed_value = ast.literal_eval(value)
        except SyntaxError as e:
            self.errorMessageWidget.showMessage(str(e))
        except ValueError:
            self.errorMessageWidget.showMessage(
                "You must enter a hexadecimal" "number preceeded by '0x' (e.g." "0xc3)."
            )
        else:
            module_state = self.module.get_state()
            module_state['memory'][parsed_adress] = parsed_value
            self.module.set_state(module_state)
            self.update_graphics_signal.emit()
