import sys

from simudator.cli.cli import CLI
from simudator.core.modules import Flag
from simudator.core.modules.register import IntegerRegister
from simudator.core.processor import Processor, Signal
from simudator.gui.gui import GUI
from simudator.processor.mia.gui import (
    AluGraphicsItem,
    ArGraphicsItem,
    AsrGraphicsItem,
    BusGraphicsItem,
    FlagGraphicsItem,
    GrxGraphicsItem,
    HrGraphicsItem,
    IrGraphicsItem,
    MiaMemoryGraphicsItem,
    MicroMemoryGraphicsItem,
    PcGraphicsItem,
    SupcGraphicsItem,
    uPcGraphicsItem,
)
from simudator.processor.mia.modules.alu import ALU
from simudator.processor.mia.modules.ar import AR
from simudator.processor.mia.modules.asr import ASR
from simudator.processor.mia.modules.bus import Bus
from simudator.processor.mia.modules.hr import HR
from simudator.processor.mia.modules.ir import IR
from simudator.processor.mia.modules.lc import LC
from simudator.processor.mia.modules.mia_grx import GRX
from simudator.processor.mia.modules.mia_memory import MiaMemory
from simudator.processor.mia.modules.micro_memory import MicroMemory
from simudator.processor.mia.modules.micro_pc import MicroPC

# Using a bare expect is generally bad code practice,
# howere here we want to load the default layout if
# anything goes wrong with loading the selected one,
# we dont care about what went wrong
from simudator.processor.mia.modules.pc import PC


class MIA_CPU(Processor):
    def __init__(self) -> None:
        super().__init__()
        # Creating all signals
        # Signals follow the naming convention 'to_from'
        self.max_line_len = 100

        # A signal used by all modules connected to the bus
        bus_control = Signal(self, "bus_control", (0, 0))

        # Dummy values used when the values in the constructor is not needed
        # unused = Signal(self)
        always_write_id = 999

        # Needed since optional arguments need to be in the correct order
        default_value = 0

        pm_bus = Signal(self, "PM-Bus")
        bus_pm = Signal(self, "Bus-PM")
        asr_bus = Signal(self, "asr_bus")
        pm_asr = Signal(self, "pm_asr")
        alu_ar = Signal(self, "alu_ar")
        ar_alu = Signal(self, "ar_alu")
        bus_ar = Signal(self, "bus_ar")
        alu_hr = Signal(self, "alu_hr")
        hr_alu = Signal(self, "hr_alu")
        bus_hr = Signal(self, "bus_hr")
        hr_bus = Signal(self, "hr_bus")
        alu_bus = Signal(self, "alu_bus")
        alu_uM = Signal(self, "alu_uM")
        z_alu = Signal(self, "z_alu")
        n_alu = Signal(self, "n_alu")
        o_alu = Signal(self, "o_alu")
        c_alu = Signal(self, "c_alu")
        grx_bus = Signal(self, "grx_bus")
        bus_grx = Signal(self, "bus_grx")
        grx_control = Signal(self, "grx_control")  # used to index registers with GRx
        m_control = Signal(self, "m_control")  # used to index registers with M
        s_control = Signal(self, "s_control")  # decides if GRx or M is used
        pc_uM = Signal(self, "pc_uM")
        bus_pc = Signal(self, "bus_pc")
        pc_bus = Signal(self, "pc_bus")
        pm_bus = Signal(self, "pm_bus")
        bus_pm = Signal(self, "bus_pm")
        always_write_signal = Signal(self, value=(999, 0))
        uPC_k1 = Signal(self, "uPC_k1")
        k1_ir = Signal(self, "k1_ir")
        uPC_k2 = Signal(self, "uPC_k2")
        ir_bus = Signal(self, "ir_bus")
        bus_ir = Signal(self, "bus_ir")
        SuPC_uPC = Signal(self, "SuPC_uPC")
        uPC_SuPC = Signal(self, "uPC_SuPC")
        uM_uPC = Signal(self, "uM_uPC")
        uPC_uM_cont = Signal(self, "uPC_uM_cont")
        uPC_uM = Signal(self, "uPC_uM")
        bus_uM = Signal(self, "bus_uM")
        lc_uM = Signal(self, "lc_uM")
        lc_uM_uADR = Signal(self, "lc_uM_uADR")
        uM_z = Signal(self, "uM_z")
        uM_n = Signal(self, "uM_n")
        uM_c = Signal(self, "uM_c")
        uM_o = Signal(self, "uM_o")
        uM_l = Signal(self, "uM_l")
        lc_bus = Signal(self, "lc_bus")
        l_lc = Signal(self, "l_lc")

        # ASR specific
        asr_bus_id = 0b111
        asr_bit_length = 8
        asr = ASR(asr_bus, pm_asr, bus_control, asr_bus_id, asr_bit_length)

        # AR specific
        bus_ar_id = 0b100
        ar_bit_length = 16
        ar = AR(ar_alu, alu_ar, bus_ar, bus_control, bus_ar_id, ar_bit_length)

        # HR specific
        hr_bus_id = 0b101
        hr_bit_length = 16
        hr = HR(alu_hr, hr_alu, bus_hr, hr_bus, bus_control, hr_bus_id, hr_bit_length)

        # ALU specific
        alu = ALU(
            "ALU",
            alu_ar,
            alu_bus,
            alu_hr,
            alu_uM,
            ar_alu,
            hr_alu,
            z_alu,
            n_alu,
            o_alu,
            c_alu,
        )

        # PC specific
        pc_bus_id = 0b011
        pc = PC(pc_uM, pc_bus, bus_pc, bus_control, pc_bus_id)

        # PM specific
        pm_size = 256
        pm_bus_id = 0b010
        pm_adress_padding = 2
        pm_value_padding = 4
        pm = MiaMemory(
            pm_bus,
            bus_pm,
            pm_asr,
            bus_control,
            pm_size,
            pm_bus_id,
            value_padding=pm_value_padding,
            adress_padding=pm_adress_padding,
            name="PM",
        )

        # GRX specific
        grx_bus_id = 0b110
        grx = GRX(
            bus_grx, grx_bus, bus_control, grx_control, m_control, s_control, grx_bus_id
        )

        # K1 specific
        k1_size = 16
        k1_adress_padding = 2
        k1_value_padding = 2
        k1 = MiaMemory(
            always_write_signal,
            uPC_k1,
            k1_ir,
            always_write_signal,
            k1_size,
            always_write_id,
            name="K1",
            adress_padding=k1_adress_padding,
            value_padding=k1_value_padding,
        )

        # K2 specific
        k2_size = 4
        k2_adress_padding = 2
        k2_value_padding = 2
        k2 = MiaMemory(
            always_write_signal,
            uPC_k2,
            m_control,
            always_write_signal,
            k2_size,
            always_write_id,
            name="K2",
            adress_padding=k2_adress_padding,
            value_padding=k2_value_padding,
        )

        # IR specific
        ir_bus_id = 0b001
        ir = IR(
            bus_ir,
            ir_bus,
            k1_ir,
            grx_control,
            m_control,
            bus_control,
            default_value,
            ir_bus_id,
        )

        # uPC specific
        uPC = MicroPC(uPC_uM_cont, uPC_k1, uPC_k2, SuPC_uPC, uPC_SuPC, uM_uPC, uPC_uM)

        # Su_PC specific
        SuPC_bit_length = 7
        SuPC_name = "SuPC"
        Su_PC = IntegerRegister(
            SuPC_uPC, uPC_SuPC, SuPC_bit_length, default_value, SuPC_name
        )

        # uM specific
        uM = MicroMemory(
            alu_uM,
            bus_control,
            bus_uM,
            s_control,
            pc_uM,
            lc_uM,
            lc_uM_uADR,
            uM_uPC,
            uPC_uM_cont,
            uPC_uM,
            uM_z,
            uM_n,
            uM_c,
            uM_o,
            uM_l,
        )

        # LC specific
        lc = LC(lc_uM, lc_bus, l_lc, lc_uM_uADR)

        # Flags
        z_flag = Flag(z_alu, uM_z, name="Z-Flag")
        n_flag = Flag(n_alu, uM_n, name="N-Flag")
        c_flag = Flag(c_alu, uM_c, name="C-Flag")
        o_flag = Flag(o_alu, uM_o, name="O-Flag")
        l_flag = Flag(l_lc, uM_l, name="L-Flag")

        # Bus
        bus = Bus(
            [bus_ir, bus_pm, bus_pc, bus_hr, bus_grx, bus_ar, bus_uM],
            [asr_bus, ir_bus, pm_bus, pc_bus, hr_bus, grx_bus, alu_bus, lc_bus],
        )

        ALLTHEMODULES = [
            uPC,
            lc,
            uM,
            Su_PC,
            ir,
            k2,
            k1,
            pc,
            ar,
            hr,
            grx,
            z_flag,
            n_flag,
            o_flag,
            l_flag,
            c_flag,
            pm,
            alu,
            asr,
            bus,
        ]

        for module in ALLTHEMODULES:
            self.add_module(module)

        self.micro_memory = uM

        self.lambdas = {}

    def should_halt(self) -> bool:
        micro_memory_state = self.micro_memory.get_state()
        return micro_memory_state["halt"]

    def run_continuously(self):
        super().run_continuously()
        # Need to do another tick  after the halt for the cycle count to be
        # correct. This is because the "halt signal" will be set at the end
        # of the clock cycle before the last clock cycle where the halt
        # technically should be executed
        if self.should_halt():
            self.do_tick()

    def launch_gui(self):
        from qtpy import QtCore
        from qtpy.QtWidgets import QApplication

        QApplication.setAttribute(QtCore.Qt.AA_EnableHighDpiScaling)
        app = QApplication(sys.argv)
        self.load_state_from_file("mia_uppg3.txt")
        self.reset()
        gui = GUI(self)

        asr = self.get_module("ASR")
        ir = self.get_module("IR")
        pm = self.get_module("PM")
        pc = self.get_module("PC")
        alu = self.get_module("ALU")
        ar = self.get_module("AR")
        hr = self.get_module("HR")
        grx = self.get_module("GRx")

        bus_signal_pairs = [
            (asr.input_s, None),
            (ir.from_bus_s, ir.to_bus_s),
            (pm.input_s, pm.output_s),
            (None, None),
            (pc.bus_input_s, pc.bus_output_s),
            (None, None),
            (alu.input_signal_b, None),
            (None, None),
            (None, ar.output_s),
            (None, None),
            (hr.input_s, hr.output_s),
            (None, None),
            (grx.from_bus, grx.to_bus),
        ]

        gui.addModuleGraphicsItem(
            BusGraphicsItem(self.get_module("Bus"), bus_signal_pairs)
        )
        gui.addModuleGraphicsItem(uPcGraphicsItem(self.get_module("uPC")))
        gui.addModuleGraphicsItem(SupcGraphicsItem(self.get_module("SuPC")))
        gui.addModuleGraphicsItem(PcGraphicsItem(self.get_module("PC")))
        gui.addModuleGraphicsItem(ArGraphicsItem(self.get_module("AR")))
        gui.addModuleGraphicsItem(AsrGraphicsItem(self.get_module("ASR")))
        gui.addModuleGraphicsItem(HrGraphicsItem(self.get_module("HR")))
        gui.addModuleGraphicsItem(MicroMemoryGraphicsItem(self.get_module("uM")))
        gui.addModuleGraphicsItem(AluGraphicsItem(self.get_module("ALU")))
        gui.addModuleGraphicsItem(GrxGraphicsItem(self.get_module("GRx")))
        gui.addModuleGraphicsItem(IrGraphicsItem(self.get_module("IR")))

        flag_modules = ["Z-Flag", "N-Flag", "C-Flag", "O-Flag", "L-Flag", "LC"]
        for name in flag_modules:
            module = self.get_module(name)
            widget = FlagGraphicsItem(module)
            gui.addModuleGraphicsItem(widget)

        memory_modules = ["PM", "K1", "K2"]
        for name in memory_modules:
            module = self.get_module(name)
            widget = MiaMemoryGraphicsItem(module)
            gui.addModuleGraphicsItem(widget)

        gui.addAllSignals()

        gui.show()
        # gui.loadLayoutFromFile("mia_layout")
        app.exec()

    def launch_cli(self):
        mia_cpu = MIA_CPU()
        cli = CLI(mia_cpu)
        cli.print_list = ["IR", "PC", "uPC"]
        mia_cpu.load_state_from_file("mia_uppg3.txt")
        mia_cpu.reset()
        cli.run()


if __name__ == "__main__":
    # this is bad terminal code, but who cares
    cpu = MIA_CPU()
    flag = None
    try:
        flag = sys.argv[1]
    except IndexError:
        pass

    if flag in ("gui", "g", "-gui"):
        cpu.launch_gui()
    else:
        cpu.launch_cli()
