from simudator.core.module import Module
from simudator.core.signal import Signal


class ALU(Module):
    """
    Module used for arithmetic logic calculations.

    Takes two input signals that are used to recive values
    for calculations.
    Takes one control signal to control what operand should
    be executed.
    Takes four flags too give extra infromation from calulations.
    Will do all calulations as two's complement binary numbers.
    """

    WORD_LENGTH = 16

    def __init__(
        self,
        name: str,
        input_signal_a: Signal,
        input_signal_b: Signal,
        input_signal_hr: Signal,
        control_signal: Signal,
        output_signal: Signal,
        output_signal_hr: Signal,
        z_signal: Signal,
        n_signal: Signal,
        o_signal: Signal,
        c_signal: Signal,
    ):
        super().__init__(name)

        # Input Signals
        self.input_signal_a = input_signal_a
        self.input_signal_b = input_signal_b
        self.control_signal = control_signal
        self.input_signal_hr = input_signal_hr

        # Add destinations
        self.input_signal_a.add_destination(self)
        self.input_signal_b.add_destination(self)
        self.control_signal.add_destination(self)
        self.input_signal_hr.add_destination(self)

        # Output Signals
        self.output_signal = output_signal
        self.output_signal_hr = output_signal_hr
        self.z_signal = z_signal
        self.n_signal = n_signal
        self.o_signal = o_signal
        self.c_signal = c_signal

        # flag values
        self.z_value = 0
        self.n_value = 0
        self.o_value = 0
        self.c_value = 0

        # internal values
        self.hr_value = None

    def update_logic(self) -> None:
        """
        Will calculate and operation depending on given contorl signal
        and update outputs with result.
        """

        value_a = self.input_signal_a.get_value()
        value_b = self.input_signal_b.get_value()
        value_hr = self.input_signal_hr.get_value()

        # bitmask control value to get first 4 bits
        control_value = self.control_signal.get_value()

        # If control value is none do nothing
        if control_value is None:
            return

        # Mask control value
        control_value = control_value & 0b1111

        # Define output value
        output_value = None

        # Check control value for action
        match control_value:
            case 0b0000:
                # Do nothing
                return

            case 0b0001:
                # set output to only b
                output_value = value_b

            case 0b0010:
                # calculate bitwise not of b
                output_value = self.ones_complement(value_b)

            case 0b0011:
                # returns 0
                output_value = 0

            case 0b0100:
                # return sum of a and b
                output_value = self.add_numbers(value_a, value_b, True)

            case 0b0101:
                # returns the differense of a and b
                output_value = self.subtract_numbers(value_a, value_b, True)

            case 0b0110:
                # returns the bitwise AND of a and b
                output_value = self.and_numbers(value_a, value_b)

            case 0b0111:
                # returns the bitwise OR of a and b
                output_value = self.or_numbers(value_a, value_b)

            case 0b1000:
                # return sum of a and b without changing flags
                output_value = self.add_numbers(value_a, value_b, False)

            case 0b1001:
                # returns the leftshifted a
                output_value = self.logical_left_shift(value_a)

            case 0b1010:
                # returns the leftshifted a (with HR)
                output_value = self.logical_left_shift(value_a, value_hr)

            case 0b1011:
                # returns the arithmetic righshifted a
                output_value = self.arithmetic_right_shift(value_a)

            case 0b1100:
                # returns the arithmetic righshifted a (with HR)
                output_value = self.arithmetic_right_shift(value_a, value_hr)

            case 0b1101:
                # returns the logically righshifted a
                output_value = self.logical_right_shift(value_a)

            case 0b1110:
                # returns a one step rotated left
                output_value = self.left_rotate(value_a)

            case 0b1111:
                # returns a one step rotated left (with HR)
                output_value = self.left_rotate(value_a, value_hr)

            case _:
                return

        # Update output signal and flags
        self.output_signal.update_value(output_value)
        self.z_signal.update_value(self.z_value)
        self.n_signal.update_value(self.n_value)
        self.o_signal.update_value(self.o_value)
        self.c_signal.update_value(self.c_value)
        self.output_signal_hr.update_value(self.hr_value)

    def print_module(self) -> None:
        print(
            "",
            self.name,
            "\n -----",
            "\n a_value: ",
            self.input_signal_a.get_value(),
            "\n b_value: ",
            self.input_signal_b.get_value(),
            "\n control_signal: ",
            self.control_signal.get_value(),
        )

    def get_input_signals(self) -> list[Signal]:
        return [
            self.input_signal_a,
            self.input_signal_b,
            self.control_signal,
            self.input_signal_hr,
        ]

    def get_output_signals(self) -> list[Signal]:
        return [
            self.output_signal,
            self.output_signal_hr,
            self.z_signal,
            self.n_signal,
            self.o_signal,
            self.c_signal,
        ]

    # ----- ALU operations -----

    def ones_complement(self, number: int) -> int:
        """
        Returns the ones complement of an int.
        """
        binary_number = self.int_to_bin(number)
        binary_complement = self.bitwise_not(binary_number)
        result = self.bin_to_int(binary_complement)
        return result

    def add_numbers(self, number_a: int, number_b: int, change_flags: bool) -> int:
        """
        Sum two numbers as binary numbers.

        Will update flags if asked to
        """

        # Convert to binary
        binary_a = self.int_to_bin(number_a)
        binary_b = self.int_to_bin(number_b)

        # Add the numbers
        result_tuple = self.add_binary_numbers(binary_a, binary_b)

        # Change flags
        if change_flags:
            self.update_flags(binary_a, binary_b, result_tuple[0], result_tuple[1])

        # Convert back to int
        result = self.bin_to_int(result_tuple[0])
        return result

    def subtract_numbers(self, number_a: int, number_b: int, change_flags: bool) -> int:
        """
        Calculates difference between to two numbers as binary numbers.

        Will update flags if asked to.
        """

        # Convert to binary
        binary_a = self.int_to_bin(number_a)
        # Convert b to a negative number
        binary_b = self.int_to_negative_bin(number_b)

        # Add the numbers
        result_tuple = self.add_binary_numbers(binary_a, binary_b)

        # Change flags
        if change_flags:
            self.update_flags(binary_a, binary_b, result_tuple[0], result_tuple[1])

        # Convert back to int
        result = self.bin_to_int(result_tuple[0])
        return result

    def and_numbers(self, number_a: int, number_b: int) -> int:
        """
        Bitwise and two int's and return the result.

        Will update flags.
        """

        # Convert to binary
        binary_a = self.int_to_bin(number_a)
        binary_b = self.int_to_bin(number_b)

        # Calulate and
        binary_and = self.binary_and(binary_a, binary_b)

        # Update flags, send data only to update z and n flags
        self.update_flags(None, None, binary_and, None)

        # Convert to int
        result = self.bin_to_int(binary_and)

        return result

    def or_numbers(self, number_a: int, number_b: int) -> int:
        """
        Bitwise or two int's and return the result.

        Will update flags.
        """
        # Convert to binary
        binary_a = self.int_to_bin(number_a)
        binary_b = self.int_to_bin(number_b)

        # Calulate and
        binary_or = self.binary_or(binary_a, binary_b)

        # Update flags, send data only to update z and n flags
        self.update_flags(None, None, binary_or, None)

        # Convert to int
        result = self.bin_to_int(binary_or)

        return result

    def logical_left_shift(self, number: int, hr_value: int = None) -> int:
        """
        Left shifts given number as a binary with word.

        Fills right with 0's after shift.
        Will update carry with removed bit.
        """

        # Convert to binary representation
        binary_number = self.int_to_bin(number)

        new_binary_number = None
        carry = None

        if hr_value is not None:
            # get hr as a binary
            hr_binary = self.int_to_bin(hr_value)

            # do the shift
            long_word_binary = hr_binary[1:] + binary_number + [0]

            # get hr number
            new_hr_number = self.bin_to_int(long_word_binary[: self.WORD_LENGTH])

            # update hr
            self.hr_value = new_hr_number

            # get new value from binary
            new_binary_number = long_word_binary[self.WORD_LENGTH :]

            # Update carry with removed bit
            carry = hr_binary[0]

        else:
            # Do shift
            new_binary_number = binary_number[1:] + [0]

            # Update carry with removed bit
            carry = binary_number[0]

        # Convert shifted to int
        value = self.bin_to_int(new_binary_number)

        self.update_flags(None, None, new_binary_number, carry)

        return value

    def logical_right_shift(self, number: int) -> int:
        """
        Left shifts given number as a binary.

        Fills right with 0's after shift.
        Will update carry with removed bit.
        """
        # Conert to binary
        binary_number = self.int_to_bin(number)

        # Do shift
        new_binary_number = [0] + binary_number[:-1]

        # Convert shifted to int
        value = self.bin_to_int(new_binary_number)

        # Update flags
        self.update_flags(None, None, new_binary_number, binary_number[-1])

        return value

    def arithmetic_right_shift(self, number: int, hr_value: int = None) -> int:
        """
        Right shifts given number as a binary with word length WORD_LENGTH.

        Fills right with most significant bit.
        Will update carry with removed bit.
        """
        # Convert to binary representation
        binary_number = self.int_to_bin(number)

        new_binary_number = None
        carry = None

        if hr_value is not None:
            # get hr as a binary
            hr_binary = self.int_to_bin(hr_value)

            # do shift
            long_word_binary = [binary_number[-1]] + hr_binary + binary_number[:-1]

            # get hr number
            new_hr_number = self.bin_to_int(long_word_binary[: self.WORD_LENGTH])

            # update hr
            self.hr_value = new_hr_number

            # get new value from binary
            new_binary_number = long_word_binary[self.WORD_LENGTH :]

            # Update carry with removed bit
            carry = binary_number[-1]

        else:
            # Do shift
            new_binary_number = [binary_number[0]] + binary_number[:-1]

            # Update carry with removed bit
            carry = binary_number[-1]

        # Convert shifted to int
        value = self.bin_to_int(new_binary_number)

        # Update flags
        self.update_flags(None, None, new_binary_number, carry)

        return value

    def left_rotate(self, number: int, hr_value: int = None) -> int:
        """
        Rotates given number left as a binary with word length WORD_LENGTH.

        Fills right with 0's after shift.
        """

        # Convert to binary representation
        binary_number = self.int_to_bin(number)

        new_binary_number = None
        carry = None

        if hr_value is not None:
            # get hr as a binary
            hr_binary = self.int_to_bin(self.input_signal_hr.get_value())

            long_word_binary = hr_binary[1:] + binary_number + [hr_binary[0]]

            # get hr number
            new_hr_number = self.bin_to_int(long_word_binary[: self.WORD_LENGTH])

            # update hr
            self.hr_value = new_hr_number

            # get new value from binary
            new_binary_number = long_word_binary[self.WORD_LENGTH :]

            # Update carry with removed bit
            carry = hr_binary[0]

        else:
            # Do rotation
            new_binary_number = binary_number[1:] + [binary_number[0]]

            # Update carry with removed bit
            carry = binary_number[0]

        # Convert shifted to int
        value = self.bin_to_int(new_binary_number)

        self.update_flags(None, None, new_binary_number, carry)

        return value

    # ----- Converters -----

    def bin_to_int(self, binary_number: [int], word_length: int = None) -> int:
        """
        Converts from binary list representation to int.

        """

        # Set default word length if none is specified
        if word_length is None:
            word_length = self.WORD_LENGTH

        number = 0

        # Get values for bits and add them
        for bit_index in range(0, word_length):
            number += binary_number[bit_index] * 2 ** (word_length - bit_index - 1)

        return number

    def int_to_bin(self, number: int, word_length: int = None) -> [int]:
        """
        Converts an int to a list binary representation where
        each index in the list represents one bit of the word.

        This is done to more easily handle edge cases that occur
        when doing arithmetic with fixed word lenghts.
        This version can not handel negative numbers.
        """

        # Set default word length if none is specified
        if word_length is None:
            word_length = self.WORD_LENGTH

        # Get list representation with correct length
        binary_number = [0 for i in range(word_length)]

        # We can now look at the other bits
        for bit_index in range(0, word_length):
            # If number is larger then bit we subtract the
            # bits decimal value from the number and make the bit 1.
            if number >= 2 ** (word_length - bit_index - 1):
                number -= 2 ** (word_length - bit_index - 1)
                binary_number[bit_index] = 1

        return binary_number

    def int_to_negative_bin(self, number: int) -> [int]:
        """
        Returns the negative of a given int as a binary list.
        """

        binary_number = self.int_to_bin(number)
        not_binary_number = self.bitwise_not(binary_number)
        binary_one = [0 for i in range(15)] + [1]
        result_tuple = self.add_binary_numbers(not_binary_number, binary_one)
        negativt_binary = result_tuple[0]
        return negativt_binary

    def int_to_bin_twocomp(self, number: int) -> [int]:
        """
        Converts an int to a list binary representation where
        each index in the list represents one bit of the word.

        This is done to more easily handle edge cases that occur
        when doing arithmetic with fixed word lenghts.
        """

        # Get list representation with correct length
        binary_number = [0 for i in range(self.WORD_LENGTH)]

        # Using twos complement representation
        # so handel negativ numbers accordingly
        if number < 0:
            binary_number[0] = 1

            # We add this so the rest of the caclulation is
            # the same for nagativ numbers
            number += 2 ** (self.WORD_LENGTH - 1)

        # We can now look at the other bits
        for bit_index in range(1, self.WORD_LENGTH):
            # If number is larger then bit we subtract the
            # bits decimal value from the number and make the bit 1.
            if number >= 2 ** (self.WORD_LENGTH - bit_index - 1):
                number -= 2 ** (self.WORD_LENGTH - bit_index - 1)
                binary_number[bit_index] = 1

        return binary_number

    def bin_to_int_twocomp(self, binary_number: [int]) -> int:
        """
        Converts from two's complement binary list representation to int.
        """
        number = 0

        # two's complement representation so most significant
        # bit will be negativ
        if binary_number[0] == 1:
            number = -(2 ** (self.WORD_LENGTH - 1))

        # Get values for other bits
        for bit_index in range(1, self.WORD_LENGTH):
            number += binary_number[bit_index] * 2 ** (self.WORD_LENGTH - bit_index - 1)

        return number

    # ----- Internal operations -----

    def update_flags(
        self, binary_a: [int], binary_b: [int], binary_result: [int], carry: int
    ) -> None:
        """
        Will update the flags depending on inputed values and the result.
        This function can take none for any input and will update the
        as well as possible for all flags.
        """

        if binary_result is not None:
            # Update z flag if we have result
            # For this we check if result is zero
            if self.bin_to_int(binary_result, len(binary_result)) == 0:
                self.z_value = 1
            else:
                self.z_value = 0

            # Update n flag if we have result
            # For this we check if the value is negativ in twos complement
            if binary_result[0] == 1:
                self.n_value = 1
            else:
                self.n_value = 0

            # If we also have given values we can calculate o flag
            if binary_a is not None and binary_b is not None:
                # Check if last bit in result is diffrent from both iven values
                if (binary_a[0] == binary_b[0]) and (binary_a[0] != binary_result[0]):
                    self.o_value = 1
                else:
                    self.o_value = 0

        # Set carry if we have it
        if carry is not None:
            self.c_value = carry

    def bitwise_not(self, binary_number: [int]) -> [int]:
        """
        Returns bitwise not of given binary number
        """

        # Make empty binary number for result
        new_binary_number = [0 for i in range(self.WORD_LENGTH)]

        # Change each bits value
        for bit_index in range(self.WORD_LENGTH):
            if binary_number[bit_index] == 0:
                new_binary_number[bit_index] = 1
            elif binary_number[bit_index] == 1:
                new_binary_number[bit_index] = 0

        return new_binary_number

    def add_binary_numbers(self, binary_a: [int], binary_b: [int]) -> ([int], int):
        """
        Returns a tuple with the sum of two binary numbers
        and the carry bit from the calculations
        """

        # Make new binary number to store results
        binary_sum = [0 for i in range(self.WORD_LENGTH)]

        # Need carry for calculation
        carry = 0

        # Do addition bit by bit and save carry between each loop
        for bit_index in range(self.WORD_LENGTH - 1, -1, -1):
            bit_sum = binary_a[bit_index] + binary_b[bit_index] + carry
            match bit_sum:
                case 0:
                    binary_sum[bit_index] = 0
                    carry = 0
                case 1:
                    binary_sum[bit_index] = 1
                    carry = 0
                case 2:
                    binary_sum[bit_index] = 0
                    carry = 1
                case 3:
                    binary_sum[bit_index] = 1
                    carry = 1

        return (binary_sum, carry)

    def binary_and(self, binary_a: int, binary_b: int) -> int:
        """
        Calculate and between binary two numbers
        """
        binary_and = [0 for i in range(self.WORD_LENGTH)]

        for binary_index in range(self.WORD_LENGTH):
            binary_and[binary_index] = binary_a[binary_index] & binary_b[binary_index]

        return binary_and

    def binary_or(self, binary_a: [int], binary_b: [int]) -> [int]:
        """
        Calculate or between two binary numbers
        """
        binary_or = [0 for i in range(self.WORD_LENGTH)]

        for binary_index in range(self.WORD_LENGTH):
            binary_or[binary_index] = binary_a[binary_index] | binary_b[binary_index]

        return binary_or
