from simudator.core.modules.memory import Memory
from simudator.core.signal import Signal
from simudator.processor.mia.modules.mia_bus_connect import MiaBusConnector


class MiaMemory(MiaBusConnector, Memory):
    """
    A MIA specific memory, functionality from the general memory
    and the MiaBusConnector is inherited. The module can be of
    arbitrary size and supports opperations that read and write
    to/from the mia bus.
    """

    def __init__(
        self,
        input_signal: Signal,
        output_signal: Signal,
        adress_signal: Signal,
        bus_control_s: Signal,
        size: int = 1,
        bus_id: int = 0,
        value_padding: int = 2,
        adress_padding: int = 2,
        name: str = "PM",
    ) -> None:
        MiaBusConnector.__init__(self, bus_control_s, bus_id)
        Memory.__init__(
            self,
            input_signal,
            output_signal,
            Signal(None),
            adress_signal,
            size,
            name=name,
            value_padding=value_padding,
            adress_padding=adress_padding,
        )
        self.bus_control_s.add_destination(self)

    def update_register(self) -> None:
        """
        Updates the value of the current adress.
        """
        if self.read_from_bus():
            self.memory[self.current_adress] = self.input_s.get_value()

    def output_register(self) -> None:
        """
        Outputs the value of the current adress.
        """
        if self.write_to_bus():
            self.output_s.update_value(self.memory[self.current_adress])
        else:
            self.output_s.update_value(None)

    def update_logic(self) -> None:
        """
        Outputs the value of the current adress unless current adress
        is None
        """
        adress_value = self.adress_s.get_value()
        if adress_value is not None:
            self.current_adress = adress_value
            if self.write_to_bus():
                self.output_s.update_value(self.memory[self.current_adress])
            else:
                self.output_s.update_value(None)

    def reset(self) -> None:
        size = len(self.memory)
        self.memory = [0 for i in range(size)]

    def load_from_str(self, state_string) -> None:
        """
        Loads the module from a string, used when loading state from
        files.
        """
        self.reset()
        lines = state_string.split("\n")
        for line in lines:
            if line:  # last line is empty from split with \n
                line_data = line.split(": ")
                adress = int(line_data[0], 16)
                value = int(line_data[1], 16)
                self.memory[adress] = value

    def get_input_signals(self) -> list[Signal]:
        return [
            self.bus_control_s,
            self.input_s,
            self.adress_s,
        ]

    def get_output_signals(self) -> list[Signal]:
        return [self.output_s]
