from simudator.core.processor import Processor
from simudator.core.signal import Signal
from simudator.processor.mia.modules.pc import PC


def test_default_name():
    cpu = Processor()
    s = Signal(cpu)
    pc = PC(s, s, s, s, s)
    assert pc.name == "PC"


def test_name():
    cpu = Processor()
    s = Signal(cpu)
    pc = PC(s, s, s, s, 0, "pc")
    assert pc.name == "pc"


def test_pc_increment():
    """
    The value of the program counter should increase by one
    when the value of the first signal (p here) is True and
    the value of the fourth is False.
    """
    cpu = Processor()

    p = Signal(cpu, "p", False)
    bus_in = Signal(cpu)
    bus_out = Signal(cpu)

    # init bus_control with different value from the bus_id
    bus_control = Signal(cpu, "test", (1, 1))

    bus_id = 4

    pc = PC(p, bus_in, bus_out, bus_control, bus_id, "PC", 0)

    cpu.add_module(pc)
    cpu.do_tick()

    p.update_value(True)
    cpu.do_tick()

    assert pc.value == 1
    cpu.do_tick()
    assert pc.value == 2
    for _ in range(5):
        cpu.do_tick()
    assert pc.value == 7


def test_set_value_from_bus():
    """
    The value of the program counter should be set by the bus when
    the value of the last signal is True and the first signal False
    """

    # init all the modules
    cpu = Processor()

    p = Signal(cpu, "p", False)
    bus_in = Signal(cpu, "bus_in", 0)
    bus_out = Signal(cpu)
    bus_control = Signal(cpu, "out", (7, 1))
    bus_id = 1

    pc = PC(p, bus_in, bus_out, bus_control, bus_id, "PC", 0)
    cpu.add_module(pc)

    # setting the destination allows the cpu to queue tick in the pc
    # when the bus recives an updated value

    bus_in.update_value(20)
    cpu.do_tick()
    assert pc.value == 20

    bus_in.update_value(10)
    cpu.do_tick()
    assert pc.value == 10


def test_write_to_bus():
    # init all the modules
    cpu = Processor()

    p = Signal(cpu, "p", False)
    bus_in = Signal(cpu, "bus_in", 1)
    bus_out = Signal(cpu, "bus_in", 2)
    bus_control = Signal(cpu, "out", (7, 1))
    bus_id = 7

    pc = PC(p, bus_in, bus_out, bus_control, bus_id, "PC", 20)
    cpu.add_module(pc)

    # setting the destination allows the cpu to queue tick in the pc
    # when the bus recives an updated value

    cpu.do_tick()
    assert bus_out.get_value() == 20

    # the program counter should not read nor write to/from bus
    bus_control.update_value((1, 2))

    cpu.do_tick()
    assert bus_out.get_value() is None


def test_exception():
    """
    The program counter should not allow the scenario where is should
    both increment the value by one and set its value from the bus.
    Below the three possible cases are tested.
    - both signals set from False to True
    - one signal already True and the other set from False to True.
    In all cases the program counter should do nothing.
    """

    # init all the modules
    cpu = Processor()

    p = Signal(cpu, "p", False)
    bus_in = Signal(cpu, "bus_in", 26)
    bus_out = Signal(cpu)
    control = Signal(cpu, "out", (2, 2))
    bus_id = 1

    pc = PC(p, bus_in, bus_out, control, bus_id, "PC", 13)
    cpu.add_module(pc)

    # set values that should raise an error
    p.update_value(True)
    control.update_value((2, 1))
    assert pc.value == 13

    # reset one of the signals, the program counter should now read
    # from the bus
    p.update_value(False)
    cpu.do_tick()
    assert pc.value == 26

    # the program counter should do nothing
    p.update_value(True)
    assert pc.value == 26

    # reset the other signal, the program counter should now increment
    # the value by one
    control.update_value((2, 2))
    cpu.do_tick()
    assert pc.value == 27

    # the program counter should do nothing
    control.update_value((2, 1))
    assert pc.value == 27


def test_get_state():
    # init all the modules
    cpu = Processor()
    p = Signal(cpu, "p", False)
    bus_in = Signal(cpu, "bus_in", 0)
    bus_out = Signal(cpu)
    control = Signal(cpu, "bus", (2, 2))
    bus_id = 1
    pc = PC(p, bus_in, bus_out, control, bus_id, "PC", 13)

    # give the cpu access to the program counter
    cpu.add_module(pc)

    # update the signals
    p.update_value(True)
    control.update_value((2, 1))

    # propagate the values to the program counter
    cpu.do_tick()

    state = pc.get_state()
    assert state["name"] == "PC"
    assert state["value"] == 13
    assert state["increment"] is True


def test_set_state():
    # init all the modules
    cpu = Processor()

    p = Signal(cpu, "p", False)
    bus_in = Signal(cpu, "bus_in", 0)
    bus_out = Signal(cpu)
    control = Signal(cpu, "bus", (2, 2))
    bus_id = 1

    pc = PC(p, bus_in, bus_out, control, bus_id, "PC", 13)

    state = {"name": "pc", "value": 23, "increment": True}
    pc.set_state(state)

    assert pc.name == "pc"
    assert pc.value == 23
    assert pc.increase_by_one is True
